#include <assert.h>
#include "Common/CommonDefs.h"
#include "CGLA/Vec3f.h"
#include "CGLA/Mat4x4f.h"
#include "Eye.h"

/****************************************************************************
 *                                                                          *
 *    (vv)                                                                  *
 *    |     /                                                               *
 *    |    /                                                                *
 *    |   / (nv)                                                            *
 *    |  /                                                                  *
 *    | /                                                                   *
 *    |/                                                                    *
 *    |----------- (uv)                                                     *
 *                                                                          *
 * nv is the line of sight, vv corresponds to the image space               *
 * y axis, and uv to the image space x axis.                                *
 *                                                                          *
 * (nv,vv,uv) are the basis of our viewing coordinate system.               *
 *                                                                          *
 ****************************************************************************/

void Eye::recalculate()
{
  //------------------------------
  // Intermediate results
  //------------------------------
      
  up.normalize();
	
  // line of sight is calculated on the 
  // basis of the eye position and 
  // focus point
      
  line_of_sight = focus_point - eye_position;


      
  //----------------------------------------------
  // The basis of our coordinate system (nv,uv,vv)
  //----------------------------------------------
      
  nv = line_of_sight;
  
  nv.normalize();  
  
//   vv = up - nv*dot(nv,up);
//   vv.normalize();
//   uv = cross(nv,vv);

  uv = cross(nv,up);
  uv.normalize();

  vv = cross(uv,nv);
  vv.normalize();

  assert(dot(vv,up)>0);


  //Create the projection Matrix
  ptm   = PerspectiveTransformation(focal_dist);
  
  // Create the transformation of the eye point

  vtm  = AffineMatrixTransformation(uv,vv,-nv);
  
  vtm &= AffineMatrixTransformation(TRANSLATE,
                                    -eye_position[0],
                                    -eye_position[1],
                                    -eye_position[2]);
  
  // Combine the twain in order to create the 
  // Viewing transformation.
  mtm = ptm & vtm;
  
}



/*****************************************************************************
 *                                                                           *
 * Eye::Eye(const Vec3f& ve, const Vec3f& vf, const int d, const Vec3f& u)*
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

Eye::Eye(const Vec3f& ve, const Vec3f& vf, const int d, const Vec3f& u):
  eye_position(ve),
  focus_point(vf),
  focal_dist(d),
  up(u)
{
  recalculate();
}
  

void Eye::set_eye_position(const Vec3f& v)
{
  eye_position = v;
  recalculate();
}

void Eye::set_focus_point(const Vec3f& v)
{
  focus_point = v;
  recalculate();
}

void Eye::set_focal_dist(const float d)
{
  focal_dist = d;
  recalculate();
}

float Eye::zoom(const float fac)
{
  focal_dist += fac;
  recalculate();
  return focal_dist;
}

  



void Eye::set_up(const Vec3f& v)
{
  up = v;
  recalculate();
}


float Eye::rotate_horiz(float a)
{
  float phi, theta, rlen;
  Vec3f r = line_of_sight * -1;
  r.get_spherical(theta, phi, rlen);
  phi += a;
  r.set_spherical(theta,phi,rlen);
  eye_position = focus_point + r;
  recalculate();
  return phi;
}

int Eye::rotate_vert(float a)
{
  float phi, theta, rlen;
  Vec3f r = line_of_sight * -1;
  r.get_spherical(theta, phi, rlen);
  theta += a;
  if (r.set_spherical(theta,phi,rlen))
    {
      eye_position = focus_point + r;
      recalculate();
      return 1;
    }
  else
    return 0;
}

float Eye::set_horiz_angle(float a)
{  
  float phi, theta, rlen;
  Vec3f r = line_of_sight * -1;
  r.get_spherical(theta, phi, rlen);
  phi = a;
  r.set_spherical(theta,phi,rlen);
  eye_position = focus_point + r;
  recalculate();
  return phi;
}


int Eye::set_vert_angle(float a)
{
  float phi, theta, rlen;
  Vec3f r = line_of_sight * -1;
  r.get_spherical(theta, phi, rlen);
  theta = a;
  if (r.set_spherical(theta,phi,rlen))
    {
      eye_position = focus_point + r;
      recalculate();
      return 1;
    }
  else
    {
      std::cout << __FILE__ << " " << __LINE__ << std::endl;
      return 0;
    }
}



