#include <algobase.h>
#include <assert.h>
#include "CGLA/Mat3x3f.h"
#include "Ellipsoid.h"


Ellipsoid::Ellipsoid(const Vec3f& _centre,
                     float _a, float _b, float _c): 
  centre(_centre), a(_a), b(_b), c(_c) 
{
  sa = sqr(a);
  sb = sqr(b);
  sc = sqr(c);

  qa = qbe(a);
  qb = qbe(b);
  qc = qbe(c);
}


float Ellipsoid::distance(const Vec3f& pos) const
{
  Vec3f pos2 = pos-centre;

  if(fabs(pos2[0]) < TINY) pos2[0] = TINY;
  if(fabs(pos2[1]) < TINY) pos2[1] = TINY;
  if(fabs(pos2[2]) < TINY) pos2[2] = TINY;
  
  float sign;
  Vec3f p0;
      
  nearest_point(pos2, p0, sign);
  
  Vec3f d = p0-pos2;
  
  assert(sign * implicit_form(pos2) >= 0);
  assert(fabs(sign) == 1);
  
  return d.length()*sign;
  
}

float Ellipsoid::foot_point(const Vec3f& p, Vec3f& g) const
{
	Vec3f p_local = p-centre;

  if(fabs(p_local[0]) < TINY) p_local[0] = TINY;
  if(fabs(p_local[1]) < TINY) p_local[1] = TINY;
  if(fabs(p_local[2]) < TINY) p_local[2] = TINY;
  
	Vec3f p0;
	float d;
	nearest_point(p_local, p0, d);

	g = p0;
	g /= Vec3f(sa, sb, sc);
	return d * ((p_local-p0).length());
}



// Equation 5 from the ``Distance to ellipsoid'' Graphics Gem by
// John Hart, Graphics Gems IV.
//
// Roots (l) of expression five are valid scaling factors which - when
// multiplied onto the gradient at a point on the ellipsoid will 
// yield a vector from that point on the ellipsoid to the point p
float Ellipsoid::eq5(const Vec3f& p, float l) const
{
  return sa*sqr(l+sb)*sqr(l+sc)*sqr(p[0])+
    sb*sqr(l+sa)*sqr(l+sc)*sqr(p[1])+
    sc*sqr(l+sb)*sqr(l+sa)*sqr(p[2])-
    sqr(l+sa)*sqr(l+sb)*sqr(l+sc);
}


// Derivative of the above function with respect to l
float Ellipsoid::deq5(const Vec3f& p, float l) const
{
  return - 2*sqr(sa+l)*sqr(sb+l)*(sc+l)
    - 2*sqr(sa+l)*(sb+l)*sqr(sc+l)
    - 2*(sa+l)*sqr(sb+l)*sqr(sc+l)
    
    + 2*sa*sqr(sb+l)*(sc+l)*sqr(p[0])
    + 2*sa*(sb+l)*sqr(sc+l)*sqr(p[0])
 
    + 2*sb*sqr(sa+l)*(sc+l)*sqr(p[1])
    + 2*sb*(sa+l)*sqr(sc+l)*sqr(p[1])

    + 2*sc*sqr(sa+l)*(sb+l)*sqr(p[2])
    + 2*sc*(sa+l)*sqr(sb+l)*sqr(p[2]);
}

/* This function finds the point on an ellipsoid which is closest
   to the point given as parameter. The method for solving this problem
   is basically to find a scaling factor such that the gradient at the 
   closest point is extended all the way to the point given as input.
   
   This involves solving a 6-degree polynomial, and the root finding method
   is Newton's method. The root finding is alleviated by the fact that we 
   know we are looking for the rightmost root, and we are therefore able
   to make a good initial guess.

   This method is due to John Hart. Graphics Gems IV
*/

void Ellipsoid::nearest_point(const Vec3f& p, 
                              Vec3f& elpsd_p, float& sign) const
{
  const int MAX_IT = 100;         // Max number of iterations

  float l; // root

  // Determine initial root guess.
  if((sign = implicit_form(p)>0 ? 1.0 : -1.0) > 0)
    l = sqrt(sqr(p[0])+sqr(p[1])+sqr(p[2]))*max(a,max(b,c));
  else
    l = 0.0;
    
  // Find point on or near ellipsoid corresponding to initial guess
  elpsd_p = Vec3f(sa*p[0]/(l+sa), sb*p[1]/(l+sb), sc*p[2]/(l+sc));

  // iteration counter
  int i=0;
  
  while(fabs(implicit_form(elpsd_p)) > 1e-7 && i<MAX_IT)
    {
      l -= eq5(p,l)/deq5(p,l);
      elpsd_p = Vec3f(sa*p[0]/(l+sa), sb*p[1]/(l+sb), sc*p[2]/(l+sc));
      i++;
    }
}


bool Ellipsoid::box_intersect(const Vec3f& Bmin,const Vec3f& Bmax) const 
{
  float dmin = 0.0;
  float dmax = 0.0;
  for(int i=0;i<3;i++)
    {
      float alpha;
      switch(i)
        {
        case 0:  alpha = a; break;
        case 1:  alpha = b; break;
        default: alpha = c; 
        }
      float da = sqr((centre[i]-Bmin[i])/alpha);
      float db = sqr((centre[i]-Bmax[i])/alpha);

      dmax += max(da,db);
      if(centre[i]<Bmin[i])
        dmin += da;
      else
        if(Bmax[i]<centre[i])
          dmin += db;
      
    }
  if (dmin<=1 && 1 <= dmax)
    return true;
  return false;
}


/** Evaluate curvature at a point.
    The method below is largely borrowed from the article 
    ``From Partial Derivatives of 3D Density Images to Ridge Lines''
    from Vis '92 
    
    There is (at least one) a weakness in the code below. The vectors
    h and f are orthogonal to g by contruction, but h becomes (0,0,0)
    when g is (0,0,g2)... This problem can be avoided by constructing
    h and f according to the direction of g.
    
    This function assumes that we are dealing with points that are on 
    the surface of the ellipsoid. The result is ill defined if this 
    assumption is false */

float Ellipsoid::curvature_priv(const Vec3f& pos) const 
{
  // Calculate the gradient.
  Vec3f g(2.0*pos[0]/sa,2.0*pos[1]/sb,2.0*pos[2]/sc);
  
  // and the normal
  Vec3f n = g;
  n.normalize();

  // Find the Hessian H. H is actually a diagonal matrix in
  // the case of an ellipsoid.
  Mat3x3f H;
  H[0][0] = 2.0/sa;
  H[1][1] = 2.0/sb;
  H[2][2] = 2.0/sc;
  
  int ax0=0;
  int ax1=1;
  int ax2=2;
 
  // We want to permute the axes so that ax2 is the one most parallel
  // to the normal, whereas ax0 and ax1 are most orthogonal.
  float max_g = -1.0;
  for (int i=0;i<3;i++)
    if (fabs(g[i])>max_g)
      {
        ax2 = i;
        max_g = fabs(g[i]);
      }
  ax0 = (ax2+1)%3;
  ax1 = (ax2+2)%3;
  
  assert(ax0 != ax1 && 
         ax1 != ax2 && 
         ax2 != ax0);
  
  // And two vectors (that form a basis together with g) which
  // by construction are of unit length and orthogonal to g.
  Vec3f h;
  h[ax0] = 1.0;
  h[ax1] = 0.0;
  h[ax2] = 0.0;  
  h -= n * dot(h,n);
  h.normalize();
  Vec3f f = cross(h, n);


  // Now find the maximum curvature
  float max_curvature = 
    ( dot(h, H * h) 
      + dot(f, H * f) 
      + sqrt(sqr(dot(h, H * h) - dot(f, H * f)) + 4*sqr(dot(h, H * f))))
    / (2*g.length());

  return max_curvature;
    
}

float Ellipsoid::curvature(const Vec3f& pos) const 
{
  return curvature_priv(pos-centre);
}


float Ellipsoid::max_curvature_in_box(const Vec3f& c0, const Vec3f& c7) const
{
  // Find the points along the edges of the box defined by c0 and c7 that 
  // intersect the ellipsoid. This involves solving a number of second 
  // degree polynomials to which there are two solutions, but due to 
  // symmetry both points must have the same curvature, and we just 
  // grab the larger value.

  Vec3f p0 = Vec3f(c0[0], c0[1], c0[2])-centre;
  Vec3f p1 = Vec3f(c7[0], c7[1], c0[2])-centre;
  Vec3f p2 = Vec3f(c0[0], c7[1], c7[2])-centre;
  Vec3f p3 = Vec3f(c7[0], c0[1], c7[2])-centre;

  Vec3f e[12];
  float  epvals[12];

  if((epvals[0]=sa*(1-sqr(p0[1])/sb-sqr(p1[2])/sc))>0)
    e[0] = Vec3f(sqrt(epvals[0]), p0[1], p0[2]);
  if((epvals[1]= sb*(1-sqr(p0[0])/sa-sqr(p0[2])/sc))>0)
    e[1] = Vec3f(p0[0], sqrt(epvals[1]), p0[2]);
  if((epvals[2]= sc*(1-sqr(p0[0])/sa-sqr(p0[1])/sb))>0)
    e[2] = Vec3f(p0[0], p0[1], sqrt(epvals[2]));

  if((epvals[3]= sa*(1-sqr(p1[1])/sb-sqr(p1[2])/sc))>0)
    e[3] = Vec3f(sqrt(epvals[3]), p1[1], p1[2]);
  if((epvals[4]= sb*(1-sqr(p1[0])/sa-sqr(p1[2])/sc))>0)
    e[4] = Vec3f(p1[0], sqrt(epvals[4]), p1[2]);

  if((epvals[5]= sc*(1-sqr(p1[0])/sa-sqr(p1[1])/sb))>0)
    e[5] = Vec3f(p1[0], p1[1], sqrt(epvals[5]));

  if((epvals[6]= sa*(1-sqr(p2[1])/sb-sqr(p2[2])/sc))>0)
    e[6] = Vec3f(sqrt(epvals[6]), p2[1], p2[2]);
  if((epvals[7]= sb*(1-sqr(p2[0])/sa-sqr(p2[2])/sc))>0)
    e[7] = Vec3f(p2[0], sqrt(epvals[7]), p2[2]);
  if((epvals[8]= sc*(1-sqr(p2[0])/sa-sqr(p2[1])/sb))>0)
    e[8] = Vec3f(p2[0], p2[1], sqrt(epvals[8]));

  if((epvals[9]= sa*(1-sqr(p3[1])/sb-sqr(p3[2])/sc))>0)
    e[9] = Vec3f(sqrt(epvals[9]), p3[1], p3[2]);
  if((epvals[10]= sb*(1-sqr(p3[0])/sa-sqr(p3[2])/sc))>0)
    e[10]= Vec3f(p3[0], sqrt(epvals[10]), p3[2]);
  if((epvals[11]= sc*(1-sqr(p3[0])/sa-sqr(p3[1])/sb))>0)
    e[11]= Vec3f(p3[0], p3[1], sqrt(epvals[11]));
  
  float max_curvature = -1.0;
  for (int i=0; i<12; i++)
    if(epvals[i]>0)
      max_curvature = max(max_curvature, curvature_priv(e[i]));
    
  return max_curvature;
 
}
  
