#ifndef __ELLIPSOID_H
#define __ELLIPSOID_H

#include "Common/CommonDefs.h"
#include "HesseNormalForm.h"
#include "CGLA/Vec3f.h"

class Ellipsoid: public HesseNormalForm
{
  Vec3f centre;
  float a,b,c;
  float qa, qb, qc;
  float sa, sb, sc;
  
private:
  
  float implicit_form(const Vec3f& pos) const
    {
      return sqr(pos[0]/a)+sqr(pos[1]/b)+sqr(pos[2]/c)-1;
    }
  
  // Equation 5 from the ``Distance to ellipsoid'' Graphics Gem by
  // John Hart, Graphics Gems IV.
  //
  // Roots (l) of expression five are valid scaling factors which - when
  // multiplied onto the gradient at a point on the ellipsoid will 
  // yield a vector from that point on the ellipsoid to the point p
  float eq5(const Vec3f& p, float l) const;
  
  // Derivative of the above function with respect to l
  float deq5(const Vec3f& p, float l) const;
  
  
  float find_alpha(const Vec3f& p) const;

  /** First parameter is the point. Upon return the second and third 
      parameters contain the nearest point and the sign of the distance
      from the original point to the nearest point on the surface */
  void nearest_point(const Vec3f& p, Vec3f&, float&) const;  

  /** Evaluate curvature at a point.
      The curvature returned is the maximum curvature.
      This function assumes that it is passed a point that has been
      transformed to ellipsoid coordinates. */
  float curvature_priv(const Vec3f&) const;
  

public:
  Ellipsoid(const Vec3f&,float,float,float);
  Ellipsoid(): a(0), b(0), c(0) {}

  /// Calculate the signed distance to the ellipsoid.
  float distance(const Vec3f&) const;

  ///
  Vec3f normal(const Vec3f& pos) const 
    {
      Vec3f pos2 = pos-centre;

      Vec3f norm(2*pos2[0]/sa, 2*pos2[1]/sb, 2*pos2[2]/sc);
      norm.normalize();
      return norm;
    }

  /// Ellipsoid box intersection by Arvo.
  bool box_intersect(const Vec3f&,const Vec3f&) const;
  
  /** Evaluate curvature at a point.
      The curvature is the maximum curvature. */
  float curvature(const Vec3f&) const;

  /// Finds first intersection with ray
  bool intersection(Ray& r) const {assert(0);return false;}

  /** Maximum curvature in a box shaped area. */
  float max_curvature_in_box(const Vec3f& c0, const Vec3f& c7) const;

	float foot_point(const Vec3f& p, Vec3f& g) const;

	bool bounding_box(Vec3f& c0, Vec3f& c7) const 
	{
		Vec3f v = Vec3f(a,b,c);
		c0 = centre - v;
		c7 = centre + v;
		return true;
	}
};


#endif
