#ifndef __RAY_H
#define __RAY_H

#include <assert.h>
#include "CGLA/Vec3f.h"

///
class Ray
{
  
private:

  /// Origin of the ray
  const Vec3f origin;

  /// Current position of ray
  Vec3f position;

  /// The normalized direction of the ray
  Vec3f direction;

  /// The parameter -i.e. the distance we have traversed along the ray
  float t;

  /// When marching this is the increment for each step.
  float increment;

  void update_pos()
    {
      position = direction;
      position *= t;
      position += origin;
    }

public:
  
  /**
    Construct a ray using three arguments:
    the position of ray starting point, the direction of ray (
    must be normalized) and the step length. */
  Ray(const Vec3f& p, const Vec3f& d, float s):
    origin(p),
    position(p),
    direction(d),
    t(0),
    increment(s){}
  
  /** Construct a ray. First argument is position. Second argument
      is the direction of the ray. The magnitude of the second argument
      is construed as the step length. */  
  Ray(const Vec3f& p, const Vec3f& d):
    origin(p),
    position(p),
    direction(d),
    t(0),
    increment(d.length())
    {
      direction.normalize();
    }
  
  /// Step along the ray. pre-increment operator.
  Vec3f operator++()
    {
      t+= increment;
      update_pos();
      return position;
    }

  ///  Step along the ray. post-increment operator.
  Vec3f operator++(int)
    {
      Vec3f tmp = position;
      t += increment;
      update_pos();
      return tmp;
    }
  
  /// Set the position of the ray
  void set_position(const Vec3f& np)
    {
      t = dot(direction,np-origin);
      update_pos();
    }

  /// Set the position of the ray
  void set_position(int axis0, float pos)
    {
			int axis1 = (axis0+1)%3;
			int axis2 = (axis1+1)%3;
			position[axis0] = pos;
			t = (position[axis0] - origin[axis0])/direction[axis0];
			position[axis1] = direction[axis1]*t + position[axis1];
			position[axis2] = direction[axis2]*t + position[axis2];
    }

  /// Change the step length
  void set_increment(const float& i) { increment = i; }

  /// 
  void set_parameter(const float& _t)
    {
      t = _t;
      position = origin + direction * t;
    }

  /// March a given distance.
  const Vec3f& march(const float& dist) 
    {
      t += dist;
      update_pos();
      return position;
    }

  /// Get ray position.
  const Vec3f& get_position() const {return position;}

  /// Get ray position.
  const Vec3f& get_origin() const {return origin;}
  
  /// Get ray parameter.
  float get_parameter() const {return t;}

  /// Get stepping length.
  float get_increment() const {return increment;}

  /// Get direction of ray.
  const Vec3f& get_direction() const {return direction;}

	/// Intersect with box.
	bool box_intersection(const Vec3f&,const Vec3f&);
};

#endif










