#ifndef __SMALLFLOAT_H
#define __SMALLFLOAT_H

#include <algobase.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <iostream>
#include "Common/CommonDefs.h"


/** The SmallFloat structure is used to represent a certain range of floating
    point numbers reasonably well. The smallest number not considered
    0 is 4.59177e-41 and the largest number is 1. 
    
    The class does no arithmetic and is normally only used in the way 
    that a normal IEEE standard 754 single precision floating point
    number may be assigned to an instance of the class and vice versa. If
    the number is bigger than 1 it is considered to be one, and if it is
    smaller than 4.59177e-41 it is considered to be 0. Some precision is 
    lost because the full mantissa is not stored (only eight bits). The
    full exponent is used, however, but numbers with exponents bigger than
    127 are not representable. Hence, no number bigger than 1 can be 
    represented. */
struct SmallFloat
{
  /// The position of the sign bit in single precision floating point
  static const int SIGN_BIT_FLT_POS = 31;

  /// The position of the exponent in single precision floating point
  static const int EXP_BIT_FLT_POS= 23;

  /// Position of the (truncated) part of the mantissa that we use.
  static const int TRUNC_MANTISSA_BIT_FLT_POS = 15;

  static const int SIGN_FLT_MASK = 0x80000000;
  static const int EXPONENT_FLT_MASK = 0x7f800000;
  static const int MANTISSA_FLT_MASK = 0x007f8000;

  /// This union is used to snatch the bits of a float.
  union FI
  {
    float x;
    unsigned int i;
  };

  /// sign of the smallfloat
  unsigned short sign: 1;
  
  /// exponent of smallfloat
  unsigned short exponent:  7;

  /// mantissa of smallfloat
  unsigned short mantissa:  8;

  /// construct zero smallfloat
  SmallFloat(): sign(0), exponent(0), mantissa(0) {}

  /// build a smallfloat from sign, exponent and mantissa
  SmallFloat(int _sign, int _exponent, int _mantissa):
    sign(_sign), exponent(_exponent), mantissa(_mantissa) {}
  
  /// Build a SmallFloat from a float
  SmallFloat(const float& x) 
    {
      (*this)=x;
    }

  /// Copy constructor.
  SmallFloat(const SmallFloat& x): 
    sign(x.sign), exponent(x.exponent), mantissa(x.mantissa) {}

  /// Assign SmallFloat to self
  void operator=(const SmallFloat& x)
    {
      sign = x.sign; 
      exponent = x.exponent;
      mantissa = x.mantissa;
    }

  /// Compare two SmallFloats
  bool operator==(const SmallFloat& x) const
    {
      if(
         sign == x.sign         &&
         exponent == x.exponent &&
         mantissa == x.mantissa )
        return true;
      return false;
    }

  /// Print the sign, exponent and mantissa.
  void print(std::ostream& os) 
    {
      os << "s = " << sign 
         << " m = " << mantissa
         << " e = " << int(exponent)
         << std::endl;
    }

  /// Operator used to assign a float to a SmallFloat
  void operator=(float f)
    {
      FI x;      
      x.x = f;
      
      sign = (x.i & SIGN_FLT_MASK) >> SIGN_BIT_FLT_POS;
      int _exponent  = ((x.i & EXPONENT_FLT_MASK) >> EXP_BIT_FLT_POS);
      if(_exponent < 127)
        {
          exponent = _exponent;
          mantissa = (x.i & MANTISSA_FLT_MASK)>>TRUNC_MANTISSA_BIT_FLT_POS;
        }
      else
        {
          exponent = 127;
          mantissa = 0;
        }
    }
  
  /// Operator used to assign the value of a SmallFloat to a float
  operator float() const
    {
      FI x;
      
      x.i = 
        int(sign)     << SIGN_BIT_FLT_POS |
        int(exponent) << EXP_BIT_FLT_POS  |
        int(mantissa) << TRUNC_MANTISSA_BIT_FLT_POS;
      
      return x.x;
    }

  /** Returns the smallest number that is not considered to
      be precisely zero. That number is 4.59177e-41. */
  static float min_val() 
    {
      FI x;
      x.i = 0x00008000;
      return x.x;
    };

  /** Returns the largest number that is not just stored as one.
      Any number greater than one is truncated to one.*/      
  static float max_val() 
    {
      FI x;
      x.i = 0x3f800000;
      return x.x;
    };

	static void test();  
};


const SmallFloat SF_ONE(0,127,0);
const SmallFloat SF_M_ONE(1,127,0);

#endif
