#include <iomanip>
#include <strstream>
#include <algorithm>

#include "CGLA/Vec3i.h"
#include "TrilinearCell.h"
#include "RootSolvers.h"
#include "Sphere.h"


using namespace std;
/// Set the point, whose value we wish to interpolate.
bool TrilinearCell::safe_set_point(const Vec3f& _point)
{
#ifndef NDEBUG
  // Initially, we don't know if we have a valid point
  // for our interpolation.
  valid_point = false;
#endif

  if(_point[0] < corner0[0]-CLOSE_TO_ZERO) {return false;}
	if(_point[0] > corner7[0]+CLOSE_TO_ZERO) {return false;}
	if(_point[1] < corner0[1]-CLOSE_TO_ZERO) {return false;}
	if(_point[1] > corner7[1]+CLOSE_TO_ZERO) {return false;}
	if(_point[2] < corner0[2]-CLOSE_TO_ZERO) {return false;}
	if(_point[2] > corner7[2]+CLOSE_TO_ZERO) {return false;}

  point[0] = max(corner0[0],min(corner7[0],_point[0]));
  point[1] = max(corner0[1],min(corner7[1],_point[1]));
  point[2] = max(corner0[2],min(corner7[2],_point[2]));

  // Calculate interpolation weights
  alpha = (point[0] - corner0[0]) / dist[0];
  beta  = (point[1] - corner0[1]) / dist[1];
  gamm  = (point[2] - corner0[2]) / dist[2];

  // Calculate one minus the weights
  m_alpha = 1.0 - alpha;
  m_beta  = 1.0 - beta;
  m_gamm  = 1.0 - gamm;

#ifndef NDEBUG          
  valid_point = true;
#endif
  return true;
}

/// Set the point, whose value we wish to interpolate.
void TrilinearCell::set_point(const Vec3f& _point)
{
  // Calculate interpolation weights
  alpha = (_point[0] - corner0[0]) / dist[0];
  beta  = (_point[1] - corner0[1]) / dist[1];
  gamm  = (_point[2] - corner0[2]) / dist[2];

  // Calculate one minus the weights
  m_alpha = 1.0 - alpha;
  m_beta  = 1.0 - beta;
  m_gamm  = 1.0 - gamm;
#ifndef NDEBUG          
  valid_point = true;
#endif

}



#define f(x)   (coeff[3]*(x)*(x)*(x) +  coeff[2]*(x)*(x) +\
                coeff[1]*(x) + coeff[0])

// // The two variables below are for heavy duty debugging.
// char nasty_str[10000];
// ostrstream* s = 0;

bool TrilinearCell::intersect(float vals[8], Ray& r)
{
  // This function finds the intersection point between a ray and 
  // the isosurface defined by the eight values at the corners of 
  // a trilinear interpolation cell. The solution is analytic 
  // rather than numerical since the maximum order of the polynomium
  // involved is 3.

  // Find ray direction in terms of cell coordinate systems
  Vec3f a[2] = {(corner7-r.get_position())/dist,
                 (r.get_position()-corner0)/dist};
  
  Vec3f b[2] = {-r.get_direction()/dist,
                 r.get_direction()/dist};

  double coeff[4] = {0,0,0,0}; // Coefficients

  // Calculate The coefficients of the cubic polynomial
  // that determines our intersection point.
  for(int k=0;k<2;k++)
    for(int j=0;j<2;j++)
      for(int i=0;i<2;i++)
        {
          int v = i+j*2+k*4;

          coeff[3] += ((b[i])[0]*(b[j])[1]*(b[k])[2]) * vals[v];
          
          coeff[2] += ((a[i])[0]*(b[j])[1]*(b[k])[2]+
                       (b[i])[0]*(a[j])[1]*(b[k])[2]+
                       (b[i])[0]*(b[j])[1]*(a[k])[2]) * vals[v];
          
          coeff[1] += ((b[i])[0]*(a[j])[1]*(a[k])[2]+
                       (a[i])[0]*(b[j])[1]*(a[k])[2]+
                       (a[i])[0]*(a[j])[1]*(b[k])[2]) * vals[v];
          
          coeff[0] += ((a[i])[0]*(a[j])[1]*(a[k])[2]) * vals[v];
        }
  
  float intersect_parameter = FLT_MAX;
  double solutions[3];
  int num;

#ifndef NDEBUG
  bool point_safely_set = safe_set_point(r.get_position());
  assert(point_safely_set);
#else
	safe_set_point(r.get_position());
#endif

	// The if below depends on what you want THINK !! DAMN
  if(interpolate(vals) < 0) return true;
	
  if (fabs(coeff[3])<1e-10)
    {
      if (fabs(coeff[2])<1e-10)
        {
          num = 1;
          solutions[0] = -coeff[0]/coeff[1];
        }
      else
        num = solve_quadric(coeff, solutions);

    }
  else
    num  = solve_cubic(coeff, solutions);
  

  assert(num<=3);
  for (int j=0;j<num;j++)
    {
      if(solutions[j] < intersect_parameter && 
         solutions[j] > 0)
        intersect_parameter = solutions[j];
    }

  
  r.march(intersect_parameter);
  bool inside = safe_set_point(r.get_position());
  return inside;

}

#undef f
